"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var _ = require("lodash");

var http = require("http");

var Client = require(".");

var _require = require("../functions"),
    formatRequest = _require.formatRequest;

var _require2 = require("../constants"),
    ERR_CODES = _require2.ERR_CODES,
    ERR_MSGS = _require2.ERR_MSGS;
/**
 * Constructor for Jsonic HTTP client
 * @class HTTPClient
 * @constructor
 * @extends Client
 * @param {Object} [options] optional settings for client
 * @return HTTPClient
 */


var HTTPClient =
/*#__PURE__*/
function (_Client) {
  _inherits(HTTPClient, _Client);

  function HTTPClient(server, options) {
    var _this;

    _classCallCheck(this, HTTPClient);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(HTTPClient).call(this, server, options)); // the content length will be calculated on a per request basis
    // according to the node http docs:
    // the encoding argument is optional and only applies when chunk is a string.
    // Defaults to 'utf8'.

    var defaults = {
      host: _this.server.host,
      port: _this.server.port || 80,
      encoding: "utf-8",
      method: "POST",
      headers: {
        "Content-Type": "application/json",
        Accept: "application/json"
      },
      path: "/"
    };
    _this.messageBuffer = "";
    _this.options = _.merge(defaults, _this.options || {});
    return _this;
  }

  _createClass(HTTPClient, [{
    key: "initClient",
    value: function initClient() {
      var _this2 = this;

      this.client = http.request(this.options, function (response) {
        _this2.writer = response;

        _this2.listen();
      });
    }
  }, {
    key: "request",
    value: function request() {
      var _this3 = this;

      return {
        message: function message(method, params) {
          var request = formatRequest({
            method: method,
            params: params,
            id: _this3.message_id,
            options: _this3.options
          });
          _this3.message_id += 1;
          return request;
        },
        send: function send(method, params) {
          return new Promise(function (resolve, reject) {
            var requestId = _this3.message_id;
            _this3.pendingCalls[requestId] = {
              resolve: resolve,
              reject: reject
            };

            var request = _this3.request().message(method, params);

            _this3.options.headers["Content-Length"] = Buffer.byteLength(request, _this3.options.encoding);

            _this3.initClient();

            _this3.client.write(request, _this3.options.encoding);

            _this3.client.end();

            _this3.client.on("error", function (error) {
              reject(error);
            });

            setTimeout(function () {
              if (_this3.pendingCalls[requestId]) {
                var error = _this3.sendError({
                  id: requestId,
                  code: ERR_CODES.timeout,
                  message: ERR_MSGS.timeout
                });

                delete _this3.pendingCalls[requestId];

                _this3.client.end();

                reject(error);
              }
            }, _this3.options.timeout);
          });
        },

        /**
         * The spec for HTTP notifications states a 204 error response with an empty body
         * Want to provide a promise response for users to know notification was received
         * or rejected by server
         */
        notify: function notify(method, params) {
          var request = formatRequest({
            method: method,
            params: params,
            options: _this3.options
          });
          return new Promise(function (resolve, reject) {
            _this3.options.headers["Content-Length"] = Buffer.byteLength(request, _this3.options.encoding);
            var notification = http.request(_this3.options, function (response) {
              if (response.statusCode === 204) {
                resolve(response);
              } else {
                reject(new Error("no response receieved for notification"));
              }
            });
            notification.write(request, _this3.options.encoding);
            notification.end();
            notification.on("error", function (error) {
              reject(error);
            });
          });
        }
      };
    }
  }, {
    key: "batch",
    value: function batch(requests) {
      var _this4 = this;

      /**
       * should receive a list of request objects
       * [client.request.message(), client.request.message()]
       * send a single request with that, server should handle it
       *
       * We want to store the IDs for the requests in the batch in an array
       * Use this to reference the batch response
       * The spec has no explaination on how to do that, so this is the solution
       */
      return new Promise(function (resolve, reject) {
        var batchIds = [];
        var batchRequests = [];
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = requests[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var _request = _step.value;
            var json = JSON.parse(_request);
            batchRequests.push(json);

            if (json.id) {
              batchIds.push(json.id);
            }
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator["return"] != null) {
              _iterator["return"]();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }

        _this4.pendingBatches[String(batchIds)] = {
          resolve: resolve,
          reject: reject
        };
        var request = JSON.stringify(batchRequests);
        _this4.options.headers["Content-Length"] = Buffer.byteLength(request, _this4.options.encoding);

        _this4.initClient();

        _this4.client.write(request, _this4.options.encoding);

        _this4.client.end();

        _this4.client.on("error", function (error) {
          reject(error);
        });

        setTimeout(function () {
          if (_this4.pendingBatches[String(batchIds)]) {
            var error = _this4.sendError({
              id: null,
              code: ERR_CODES.timeout,
              message: ERR_MSGS.timeout
            });

            delete _this4.pendingBatches[String(batchIds)];
            reject(error);
          }
        }, _this4.options.timeout);

        _this4.on("batchResponse", function (batch) {
          var batchResponseIds = [];
          batch.forEach(function (message) {
            if (message.id) {
              batchResponseIds.push(message.id);
            }
          });

          if (_.isEmpty(batchResponseIds)) {
            resolve([]);
          }

          var _loop = function _loop() {
            var ids = _Object$keys[_i];

            if (_.isEmpty(_.difference(JSON.parse("[".concat(ids, "]")), batchResponseIds))) {
              var response = _objectSpread({
                body: batch
              }, _this4.writer);

              batch.forEach(function (message) {
                if (message.error) {
                  // reject the whole message if there are any errors
                  if (_this4.pendingBatches[ids] !== undefined) {
                    _this4.pendingBatches[ids].reject(response);

                    delete _this4.pendingBatches[ids];
                  }
                }
              });

              if (_this4.pendingBatches[ids] !== undefined) {
                _this4.pendingBatches[ids].resolve(response);

                delete _this4.pendingBatches[ids];
              }
            }
          };

          for (var _i = 0, _Object$keys = Object.keys(_this4.pendingBatches); _i < _Object$keys.length; _i++) {
            _loop();
          }
        });

        _this4.on("batchError", function (error) {
          reject(error);
        });
      });
    }
  }]);

  return HTTPClient;
}(Client);

module.exports = HTTPClient;