"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var EventEmitter = require("events");

var _ = require("lodash");

var http = require("http");

var _require = require("../constants"),
    ERR_CODES = _require.ERR_CODES,
    ERR_MSGS = _require.ERR_MSGS;
/**
 * @class Client
 * @extends require('events').EventEmitter
 * @param {Object} [server] object with host and port of server
 * @param {Object} [options]
 * @param {Number} [options.version=2] JSON-RPC version to use (1|2)
 * @param {String} [options.delimiter="\n"] delimiter to use for requests
 * @param {Number} [options.timeout=30] timeout for request
 * @return {Client}
 */


var Client =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Client, _EventEmitter);

  function Client(options) {
    var _this;

    _classCallCheck(this, Client);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Client).call(this));

    if (!(_assertThisInitialized(_this) instanceof Client)) {
      return _possibleConstructorReturn(_this, new Client(options));
    }

    var defaults = {
      host: "127.0.0.1",
      port: "8100",
      version: "2.0",
      delimiter: "\n",
      timeout: 30,
      retries: 2
    };
    _this.client = undefined;
    _this.message_id = 1;
    _this.serving_message_id = 1;
    _this.pendingCalls = {};
    _this.pendingBatches = {};
    _this.attached = false;
    /**
     * we can receive whole messages, or parital so we need to buffer
     *
     * whole message: {"jsonrpc": 2.0, "params": ["hello"], id: 1}
     *
     * partial message: {"jsonrpc": 2.0, "params"
     */

    _this.messageBuffer = "";
    _this.responseQueue = {};
    _this.options = _.merge(defaults, options || {});
    _this.options.timeout = _this.options.timeout * 1000;
    var _this$options = _this.options,
        host = _this$options.host,
        port = _this$options.port;
    _this.server = {
      host: host,
      port: port
    };
    return _this;
  } // http client doesnt need this method


  _createClass(Client, [{
    key: "connect",
    value: function connect() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "end",
    value: function end() {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        _this2.client.end(function (error) {
          if (error) {
            reject();
          }

          resolve();
        });
      });
    }
  }, {
    key: "request",
    value: function request() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "batch",
    value: function batch() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "notify",
    value: function notify() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "subscribe",
    value: function subscribe() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "handleResponse",
    value: function handleResponse(id) {
      if (!(this.pendingCalls[id] === undefined)) {
        var response = this.responseQueue[id];

        if (this.writer instanceof http.IncomingMessage) {
          // want to allow users to access the headers, status code etc.
          response = _objectSpread({
            body: this.responseQueue[id]
          }, this.writer);
        }

        this.pendingCalls[id].resolve(response);
        delete this.responseQueue[id];
      }
    }
  }, {
    key: "verifyData",
    value: function verifyData(messages) {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = messages[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var chunk = _step.value;

          try {
            // sometimes split messages have empty string at the end
            // just ignore it
            if (chunk !== "") {
              // will throw an error if not valid json
              var message = JSON.parse(chunk);

              if (_.isArray(message)) {
                // possible batch request
                try {
                  this.emit("batchResponse", message);
                } catch (e) {
                  var error = this.sendError({
                    id: null,
                    code: ERR_CODES.parseError,
                    message: ERR_MSGS.parseError
                  });
                  this.emit("batchError", error);
                }
              } else if (!_.isObject(message)) {
                // error out if it cant be parsed
                var _error = this.sendError({
                  id: null,
                  code: ERR_CODES.parseError,
                  message: ERR_MSGS.parseError
                });

                this.handleError(_error);
              } else if (!message.id) {
                // special case http response
                // may want to still know the body
                // this is not in spec at all
                if (this.writer instanceof http.IncomingMessage) {
                  var _error2 = this.sendError({
                    id: this.serving_message_id,
                    code: ERR_CODES.parseError,
                    message: ERR_MSGS.parseError
                  });

                  this.writer.response = message;
                  this.handleError(_error2);
                } // no id, so assume notification


                this.emit("notify", message);
              } else if (message.error) {
                // got an error back so reject the message
                var _error3 = this.sendError({
                  jsonrpc: message.jsonrpc,
                  id: message.id,
                  code: message.error.code,
                  message: message.error.message
                });

                this.handleError(_error3);
              } else if (!message.method) {
                // no method, so assume response
                this.serving_message_id = message.id;
                this.responseQueue[this.serving_message_id] = message;
                this.handleResponse(this.serving_message_id);
              } else {
                throw new Error();
              }
            }
          } catch (e) {
            if (e instanceof TypeError) {
              var _error4 = this.sendError({
                id: this.serving_message_id,
                code: ERR_CODES.parseError,
                message: ERR_MSGS.parseError
              });

              this.handleError(_error4);
            } else {
              var _error5 = this.sendError({
                id: this.serving_message_id,
                code: ERR_CODES.internal,
                message: ERR_MSGS.internal
              });

              this.handleError(_error5);
            }
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator["return"] != null) {
            _iterator["return"]();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }
  }, {
    key: "listen",
    value: function listen() {
      var _this3 = this;

      this.writer.on("data", function (data) {
        _this3.messageBuffer += data;

        if (_this3.writer instanceof http.IncomingMessage) {
          _this3.writer.on("end", function () {
            // only expect one response per request for http
            var messages = _this3.messageBuffer;
            _this3.messageBuffer = "";

            _this3.verifyData(Array(messages));
          });
        } else {
          var messages = _this3.messageBuffer.split(_this3.options.delimiter);

          if (messages.length > 1) {
            // otherwise messages are still coming in and request
            // possibly hasnt finished
            _this3.messageBuffer = "";

            _this3.verifyData(messages);
          }
        }
      });
      this.client.on("close", function () {
        _this3.attached = false;

        _this3.client.removeAllListeners();

        _this3.emit("serverDisconnected");
      });
    }
  }, {
    key: "serverDisconnected",
    value: function serverDisconnected(cb) {
      this.on("serverDisconnected", function () {
        return cb();
      });
    }
  }, {
    key: "handleError",
    value: function handleError(error) {
      var response = error;

      if (this.writer instanceof http.IncomingMessage) {
        // want to allow users to access the headers, status code etc.
        response = _objectSpread({
          body: error
        }, this.writer);
      }

      this.pendingCalls[error.id].reject(response);
    }
  }, {
    key: "sendError",
    value: function sendError(_ref) {
      var jsonrpc = _ref.jsonrpc,
          id = _ref.id,
          code = _ref.code,
          message = _ref.message;
      var response;

      if (this.options.version === "2.0") {
        response = {
          jsonrpc: jsonrpc || this.options.version,
          error: {
            code: code,
            message: message || "Unknown Error"
          },
          id: id
        };
      } else {
        response = {
          result: null,
          error: {
            code: code,
            message: message || "Unknown Error"
          },
          id: id
        };
      }

      return response;
    }
  }]);

  return Client;
}(EventEmitter);

module.exports = Client;
/**
 * TCP client constructor
 * @type ClientTcp
 * @static
 */

Client.tcp = require("./tcp");
/**
 * HTTP client constructor
 * @type ClientHTTP
 * @static
 */

Client.http = require("./http");