"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var net = require("net");

var _ = require("lodash");

var Client = require(".");

var _require = require("../functions"),
    formatRequest = _require.formatRequest;

var _require2 = require("../constants"),
    ERR_CODES = _require2.ERR_CODES,
    ERR_MSGS = _require2.ERR_MSGS;
/**
 * Constructor for Jsonic TCP client
 * @class TCPClient
 * @constructor
 * @extends Client
 * @param {Object} [options] optional settings for client
 * @return TCPClient
 */


var TCPClient =
/*#__PURE__*/
function (_Client) {
  _inherits(TCPClient, _Client);

  function TCPClient() {
    _classCallCheck(this, TCPClient);

    return _possibleConstructorReturn(this, _getPrototypeOf(TCPClient).apply(this, arguments));
  }

  _createClass(TCPClient, [{
    key: "connect",
    value: function connect() {
      var _this = this;

      return new Promise(function (resolve, reject) {
        if (_this.attached) {
          // not having this caused MaxEventListeners error
          return reject(Error("client already connected"));
        }

        _this.client = new net.Socket();

        _this.client.connect(_this.server);

        _this.client.setEncoding("utf8");

        _this.client.on("connect", function () {
          _this.attached = true;
          _this.writer = _this.client; // start listeners, response handlers and error handlers

          _this.listen();

          resolve(_this.server);
        });

        var retries = _this.options.retries;

        _this.client.on("error", function (error) {
          if (error.code === "ECONNREFUSED" && retries) {
            retries -= 1;
            process.stdout.write("Unable to connect. Retrying. ".concat(retries, " attempts left.\n"));
            setTimeout(function () {
              _this.client.connect(_this.server);
            }, 5000);
          } else {
            reject(error);
          }
        });
      });
    }
  }, {
    key: "request",
    value: function request() {
      var _this2 = this;

      return {
        message: function message(method, params) {
          var id = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;
          var request = formatRequest({
            method: method,
            params: params,
            id: id ? _this2.message_id : undefined,
            options: _this2.options
          });
          _this2.message_id += 1;
          return request;
        },
        send: function send(method, params) {
          return new Promise(function (resolve, reject) {
            var requestId = _this2.message_id;
            _this2.pendingCalls[requestId] = {
              resolve: resolve,
              reject: reject
            };

            try {
              _this2.client.write(_this2.request().message(method, params));
            } catch (e) {
              if (e instanceof TypeError) {
                // this.client is probably undefined
                reject(new Error("Unable to send request. ".concat(e.message)));
              }
            }

            setTimeout(function () {
              if (_this2.pendingCalls[requestId]) {
                var error = _this2.sendError({
                  id: requestId,
                  code: ERR_CODES.timeout,
                  message: ERR_MSGS.timeout
                });

                delete _this2.pendingCalls[requestId];
                reject(error);
              }
            }, _this2.options.timeout);
          });
        },
        notify: function notify(method, params) {
          var request = formatRequest({
            method: method,
            params: params,
            options: _this2.options
          });
          return new Promise(function (resolve, reject) {
            try {
              _this2.client.write(request, function () {
                resolve("notification sent");
              });
            } catch (e) {
              if (e instanceof TypeError) {
                // this.client is probably undefined
                reject(new Error("Unable to send request. ".concat(e.message)));
              }
            }
          });
        }
      };
    }
  }, {
    key: "batch",
    value: function batch(requests) {
      var _this3 = this;

      /**
       * should receive a list of request objects
       * [client.request.message(), client.request.message()]
       * send a single request with that, server should handle it
       *
       * We want to store the IDs for the requests in the batch in an array
       * Use this to reference the batch response
       * The spec has no explaination on how to do that, so this is the solution
       */
      return new Promise(function (resolve, reject) {
        var batchIds = [];
        var batchRequests = [];
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = requests[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var _request = _step.value;
            var json = JSON.parse(_request);
            batchRequests.push(json);

            if (json.id) {
              batchIds.push(json.id);
            }
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator["return"] != null) {
              _iterator["return"]();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }

        _this3.pendingBatches[String(batchIds)] = {
          resolve: resolve,
          reject: reject
        };
        var request = JSON.stringify(batchRequests);

        try {
          _this3.client.write(request + _this3.options.delimiter);
        } catch (e) {
          if (e instanceof TypeError) {
            // this.client is probably undefined
            reject(new Error("Unable to send request. ".concat(e.message)));
          }
        }

        setTimeout(function () {
          if (_this3.pendingBatches[String(batchIds)]) {
            var error = _this3.sendError({
              id: null,
              code: ERR_CODES.timeout,
              message: ERR_MSGS.timeout
            });

            delete _this3.pendingBatches[String(batchIds)];
            reject(error);
          }
        }, _this3.options.timeout);

        _this3.on("batchResponse", function (batch) {
          var batchResponseIds = [];
          batch.forEach(function (message) {
            if (message.id) {
              batchResponseIds.push(message.id);
            }
          });

          if (_.isEmpty(batchResponseIds)) {
            resolve([]);
          }

          var _loop = function _loop() {
            var ids = _Object$keys[_i];

            if (_.isEmpty(_.difference(JSON.parse("[".concat(ids, "]")), batchResponseIds))) {
              batch.forEach(function (message) {
                if (message.error) {
                  // reject the whole message if there are any errors
                  if (_this3.pendingBatches[ids] !== undefined) {
                    _this3.pendingBatches[ids].reject(batch);

                    delete _this3.pendingBatches[ids];
                  }
                }
              });

              if (_this3.pendingBatches[ids] !== undefined) {
                _this3.pendingBatches[ids].resolve(batch);

                delete _this3.pendingBatches[ids];
              }
            }
          };

          for (var _i = 0, _Object$keys = Object.keys(_this3.pendingBatches); _i < _Object$keys.length; _i++) {
            _loop();
          }
        });

        _this3.on("batchError", function (error) {
          reject(error);
        });
      });
    }
    /**
     * @params {String} [method] method to subscribe to
     * @params {Function} [cb] callback function to invoke on notify
     */

  }, {
    key: "subscribe",
    value: function subscribe(method, cb) {
      this.on("notify", function (message) {
        try {
          if (message.method === method) {
            return cb(undefined, message);
          }
        } catch (e) {
          return cb(e);
        }
      });
    }
  }]);

  return TCPClient;
}(Client);

module.exports = TCPClient;