"use strict";

var isString = require("lodash/isString");

var isUndefined = require("lodash/isUndefined");

var isObject = require("lodash/isObject");

var isArray = require("lodash/isArray");

var formatRequest = function formatRequest(_ref) {
  var method = _ref.method,
      params = _ref.params,
      id = _ref.id,
      options = _ref.options;

  if (!isString(method)) {
    throw new TypeError("".concat(method, " must be a string"));
  }

  var request = {
    method: method
  }; // assume 2.0 request unless otherwise specified

  if (!options.version || options.version !== 1) {
    request.jsonrpc = "2.0";
  }

  if (!params) {
    throw new Error("params must be defined");
  }

  if (params) {
    if (!isObject(params) && !isArray(params)) {
      throw new TypeError("".concat(params, " must be an object or array"));
    }

    request.params = params;
  } // assume notification otherwise


  if (!isUndefined(id)) {
    request.id = id;
  }

  var messageString = JSON.stringify(request) + options.delimiter;
  return messageString;
};

var formatResponse = function formatResponse(_ref2) {
  var jsonrpc = _ref2.jsonrpc,
      id = _ref2.id,
      method = _ref2.method,
      result = _ref2.result,
      params = _ref2.params;
  var response = {};
  response.result = result;

  if (params) {
    response.params = params;
  }

  if (!jsonrpc || jsonrpc === 1) {
    // 1.0 response
    response.error = null;
  } else {
    // 2.0 response, dont include null error and include jsonrpc version
    response.jsonrpc = "2.0";
  }

  if (id) {
    response.id = id;
  } // could be notification


  if (method && !id) {
    response.method = method;
  }

  return JSON.stringify(response);
};

module.exports = {
  formatRequest: formatRequest,
  formatResponse: formatResponse
};