"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var EventEmitter = require("events");

var _ = require("lodash");

var isObject = require("lodash/isObject");

var isArray = require("lodash/isArray");

var _require = require("../functions"),
    formatResponse = _require.formatResponse;

var _require2 = require("../constants"),
    ERR_CODES = _require2.ERR_CODES,
    ERR_MSGS = _require2.ERR_MSGS;
/**
 * @class Server
 * @extends require('events').EventEmitter
 * @param {Object} [host] host IP to connect with
 * @param {Object} [host] host port to connect with
 * @param {Object} [options]
 * @param {Number} [options.version=2] JSON-RPC version to use (1|2)
 * @param {String} [options.delimiter="\n"] delimiter to use for requests
 * @param {Boolean} [options.exlusive=false] disallow port sharing
 * @return {Client}
 */


var Server =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Server, _EventEmitter);

  function Server(options) {
    var _this;

    _classCallCheck(this, Server);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Server).call(this));

    if (!(_assertThisInitialized(_this) instanceof Server)) {
      return _possibleConstructorReturn(_this, new Server(options));
    }

    var defaults = {
      host: "localhost",
      port: 8100,
      exclusive: false,
      version: "2.0",
      delimiter: "\n"
    };
    _this.options = _.merge(defaults, options || {});
    _this.messageBuffer = "";
    _this.methods = {};
    _this.listening = false;
    return _this;
  }

  _createClass(Server, [{
    key: "listen",
    value: function listen() {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        if (_this2.listening) {
          // not having this caused MaxEventListeners error
          reject(Error("server already listening"));
        }

        var _this2$options = _this2.options,
            host = _this2$options.host,
            port = _this2$options.port,
            exclusive = _this2$options.exclusive;

        _this2.server.listen({
          host: host,
          port: port,
          exclusive: exclusive
        });

        _this2.server.on("listening", function () {
          _this2.listening = true;

          _this2.handleData();

          _this2.handleError();

          resolve({
            host: _this2.server.address().address,
            port: _this2.server.address().port
          });
        });

        _this2.server.on("error", function (error) {
          _this2.listening = false;
          reject(error);
        });

        _this2.server.on("close", function () {
          _this2.listening = false;
        });
      });
    }
  }, {
    key: "close",
    value: function close() {
      var _this3 = this;

      this.listening = false;
      return new Promise(function (resolve, reject) {
        _this3.server.close(function (error) {
          if (error) {
            reject();
          }

          resolve();
        });
      });
    }
  }, {
    key: "method",
    value: function method(name, cb) {
      // add the method and its associated callback to the object
      this.methods[name] = cb;
    }
  }, {
    key: "onNotify",
    value: function onNotify(method, cb) {
      this.on("notify", function (message) {
        try {
          if (message.method === method) {
            return cb(undefined, message);
          }
        } catch (e) {
          return cb(e);
        }
      });
    }
  }, {
    key: "handleData",
    value: function handleData() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "handleBatchRequest",
    value: function handleBatchRequest(batch) {
      var _this4 = this;

      return new Promise(function (resolve, reject) {
        var requests = batch;
        var batchRequests = requests.map(function (request) {
          return _this4.validateMessage(request).then(function (message) {
            return _this4.getResult(message).then(function (result) {
              return JSON.parse(result);
            })["catch"](function (error) {
              throw error;
            });
          })["catch"](function (error) {
            throw error;
          });
        });
        Promise.all(batchRequests.map(function (promise) {
          return promise["catch"](function (error) {
            return error;
          });
        })).then(function (result) {
          resolve(result);
        })["catch"](function (error) {
          reject(error);
        });
      });
    }
  }, {
    key: "validateRequest",
    value: function validateRequest(request) {
      var _this5 = this;

      return new Promise(function (resolve, reject) {
        try {
          var message = JSON.parse(request);
          resolve(message);
        } catch (e) {
          reject(_this5.sendError(null, ERR_CODES.parseError, ERR_MSGS.parseError));
        }
      });
    }
  }, {
    key: "validateMessage",
    value: function validateMessage(message) {
      var _this6 = this;

      return new Promise(function (resolve, reject) {
        if (_.isArray(message)) {
          // possible batch request
          if (_.isEmpty(message)) {
            var error = _this6.sendError(null, ERR_CODES.invalidRequest, ERR_MSGS.invalidRequest);

            return reject(error);
          }

          return _this6.handleBatchRequest(message).then(function (responses) {
            resolve({
              batch: responses
            });
          })["catch"](function (error) {
            reject(JSON.stringify(error));
          });
        }

        if (!_.isObject(message)) {
          reject(_this6.sendError(null, ERR_CODES.invalidRequest, ERR_MSGS.invalidRequest));
        }

        if (!_.isString(message.method)) {
          reject(_this6.sendError(message.id, ERR_CODES.invalidRequest, ERR_MSGS.invalidRequest));
        }

        if (!message.id) {
          // no id, so assume notification
          resolve({
            notification: message
          });
        }

        if (message.jsonrpc) {
          if (_this6.options.version !== "2.0") {
            reject(_this6.sendError(message.id, ERR_CODES.invalidRequest, ERR_MSGS.invalidRequest));
          }
        }

        if (!_this6.methods[message.method]) {
          reject(_this6.sendError(message.id, ERR_CODES.methodNotFound, ERR_MSGS.methodNotFound));
        }

        if (!isArray(message.params) && !isObject(message.params)) {
          reject(_this6.sendError(message.id, ERR_CODES.invalidParams, ERR_MSGS.invalidParams));
        } // data looks good


        resolve(message);
      });
    }
  }, {
    key: "handleValidation",
    value: function handleValidation(chunk) {
      var _this7 = this;

      var validRequest = this.validateRequest(chunk).then(function (result) {
        return result;
      })["catch"](function (error) {
        throw error;
      });
      var validMessage = validRequest.then(function (result) {
        return _this7.validateMessage(result).then(function (message) {
          return message;
        })["catch"](function (error) {
          throw error;
        });
      })["catch"](function (error) {
        throw error;
      });
      return [validRequest, validMessage];
    }
  }, {
    key: "getResult",
    value: function getResult(message) {
      var _this8 = this;

      var params = message.params;
      return new Promise(function (resolve, reject) {
        try {
          var result = _this8.methods[message.method](params);

          var response = formatResponse({
            jsonrpc: message.jsonrpc,
            id: message.id,
            result: result
          });
          resolve(response);
        } catch (e) {
          var error = _this8.sendError(message.id, ERR_CODES.internal);

          if (e instanceof TypeError) {
            error = _this8.sendError(message.id, ERR_CODES.invalidParams, ERR_MSGS.invalidParams);
          }

          reject(error);
        }
      });
    }
  }, {
    key: "clientConnected",
    value: function clientConnected() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "clientDisconnected",
    value: function clientDisconnected() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "handleError",
    value: function handleError() {
      var _this9 = this;

      this.on("error", function (error) {
        _this9.listening = false;
        return error;
      });
    }
  }, {
    key: "sendError",
    value: function sendError(id, code) {
      var message = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
      var response;

      if (this.options.version === "2.0") {
        response = {
          jsonrpc: this.options.version,
          error: {
            code: code,
            message: message || "Unknown Error"
          },
          id: id
        };
      } else {
        response = {
          result: null,
          error: {
            code: code,
            message: message || "Unknown Error"
          },
          id: id
        };
      }

      return response;
    }
  }]);

  return Server;
}(EventEmitter);

module.exports = Server;
/**
 * HTTP server constructor
 * @type ServerHTTP
 * @static
 */

Server.http = require("./http");
/**
 * TCP server constructor
 * @type ServerTCP
 * @static
 */

Server.tcp = require("./tcp");
/**
 * WS server constructor
 * @type ServerWS
 * @static
 */

Server.ws = require("./ws");